#include "DOAContext.h"

#include "OCD_include.h"
#include "Proxy.h"
#include "DOAStateInfo.h"
#include "DOAOCD.h"
#include "DOAGlobals.h"
#include "ERecord.h"
#include "DOAPacket.h"

#include "Resolve/DOADNS.h"
#include "Resolve/DOADHT.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
 
DOAContext::DOAContext ( DOAOCD* doaOCD, string fileName )
{
	this->doaOCD = doaOCD;	
	
	/////////////////////////////////////
	// initialize values from config file
	/////////////////////////////////////
	
	ConfigFile myConfigFile (fileName, "DOAOCDConfigFile");

	//### Read the number of EID delegates for this DOA from config file ###
	myConfigFile.readUShortAttrib ("//DOAOCDConfigFile", "numEIDDelegates", 
				       &numEIDDelegates,
				       true);

	if(numEIDDelegates > 0){
	  //### Read the delegate EIDs for this DOA from config file ###
	  char tempAddress[1000];
	  memset(tempAddress, 0x00, 1000);
	  myConfigFile.readStringAttrib ("//DOAOCDConfigFile", "EIDDelegates", 
				       tempAddress,
				       true);
	  EIDDelegates = new char[strlen(tempAddress)+1];
	  memcpy(EIDDelegates,tempAddress,strlen(tempAddress)+1);
	}

	//### Read the number of IP delegates for this DOA from config file ###
	myConfigFile.readUShortAttrib ("//DOAOCDConfigFile", "numIPDelegates", 
				       &numIPDelegates,
				       true);

	if(numIPDelegates > 0){
	  //### Read the delegate EIDs for this DOA from config file ###
	  char tempAddress[200];
	  memset(tempAddress, 0x00, 200);
	  myConfigFile.readStringAttrib ("//DOAOCDConfigFile", "IPDelegates", 
				       tempAddress,
				       true);
	  IPDelegates = new char[strlen(tempAddress)+1];
	  memcpy(IPDelegates,tempAddress,strlen(tempAddress)+1);
	}

	//### Read the doa format address for this DOA from config file ###
	char tempAddress[200];
	memset(tempAddress, 0x00, 200);
	myConfigFile.readStringAttrib ("//DOAOCDConfigFile", "myAddress", 
				       tempAddress,
				       true);
	myAddress = new char[strlen(tempAddress)+1];
	memcpy(myAddress,tempAddress,strlen(tempAddress)+1);

	//### Read the doa IP address for this DOA from config file ###
	memset(tempAddress, 0x00, 200);
	myConfigFile.readStringAttrib ("//DOAOCDConfigFile", "myIP", 
				       tempAddress,
				       true);
	myIP = new char[strlen(tempAddress)+1];
	memcpy(myIP,tempAddress,strlen(tempAddress)+1);

	//### Read the EID for this DOA from config file ###
	char tempEID[EID_SIZE+1];
	myConfigFile.readStringAttrib ("//DOAOCDConfigFile", "myEID",
				       tempEID,
				       true);
	memcpy(&mEID,tempEID,20);
	tempEID[EID_SIZE] = '\0';

	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in numEIDDelegates from CONFIGFILE: %d**\n", numEIDDelegates);
	if(numEIDDelegates > 0) {
	  PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in delegate EIDs from CONFIGFILE: %s**\n", EIDDelegates);
	}
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in numIPDelegates from CONFIGFILE: %d**\n", numIPDelegates);
	if(numIPDelegates > 0) {
	  PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in delegate IPs from CONFIGFILE: %s**\n", IPDelegates);
	}
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in DOA Address from CONFIGFILE: %s**\n", myAddress);
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in IP Address from CONFIGFILE: %s**\n", myIP);
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "\n**Read in EID from CONFIGFILE: %s**\n", tempEID);
}

void DOAContext::init(){
	
    // Open the socket on which DOA listens for packets.
    mySockFD =  socket (AF_INET, SOCK_DGRAM, 0); 
    
    // bind the socket to myPortNum
    struct sockaddr_in my_addr;    // my address information
    my_addr.sin_family = AF_INET;         // host byte order
    my_addr.sin_port = htons (DOA_PORT);  // short, network byte order
    my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
    memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct

    if (bind(mySockFD, (struct sockaddr *)&my_addr,
			sizeof(struct sockaddr)) == -1) {
        PRINT_INFO (
                    INFO_LEVEL_FATAL_ERROR,
                    "DOA OCD: Unable to initialize: %s.\n",
                    strerror (errno)
                );
        EXIT_ON_ERROR;

        // Note that this will exit the whole proxy.  Even other OCDs
        // that are plugged into the proxy will be terminated.
        // This is not a polite thing to do.
    }

    // We have bound the socket to the DOA_PORT.
    // Now, add a callback to be invoked when a packet is received 
    // on the socket.
    TObjectFunctor<DOAContext> *fdfunc = new TObjectFunctor<DOAContext>(this, &DOAContext::receiveDOA);
   	doaOCD->proxy->addFDCallback(mySockFD, fdfunc);
   	

	/////////////////////
	// WITH SECURITY
	/////////////////////
	
	// generate private/public key seeded by IP address
	// generate EID by hashing the public key
	// register with DNS
	// create ERecord
	// sign ERecord with private key
	// register the ERecord/cert with the DHT

	/////////////////////
	// WITHOUT SECURITY
	/////////////////////
       
	// for now, eid initialized from config file.  No signing/verification.

	// register with DNS (DOAAddress (e.g. "this_host.doa") -> EID)
	registerWithDNS(this);

	// registe with DHT (eid -> erecord)
	registerWithDHT(this);
}

int DOAContext::setupTunnel(DOAStateInfo* ocdState){
	    
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"setupTunnel...\n");
    
    // Strip the .doa from the url and send it as a DNS request
	char tmpEndPointName [MAX_URL_LEN];
    strncpy (tmpEndPointName, ocdState->endPointName, MAX_URL_LEN);
    stripSuffix (tmpEndPointName, ".doa");
   
   	return doaDNSRequest(tmpEndPointName,ocdState)?1:0;
}

/* Sends a DOA packet by first constructing a DOA packet and then
 * sending it out to the socket */
int DOAContext::sendDOA(DOAStateInfo* info, char* data, int len){
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"sendDOA...\n");

	// if the StateInfo isn't done resolving DHT stuff, queue the packet to be sent later
	if (!info->isReady()) {
		// if the StateInfo isn't resolving the address, set off the resolution cascade
		if (!info->isResolving()) {
			info->checkAndResolve();
		}
		info->enqueuePacket(data, len);
		return info->numQPackets;
	}

	DOAPacket packet = DOAPacket(info,this,data,len);
	
	struct sockaddr_in tmpDestAddr;
    tmpDestAddr.sin_addr = info->dest_ip;
    tmpDestAddr.sin_port = htons(DOA_PORT);
   
    tmpDestAddr.sin_family = AF_INET;
    memset(&(tmpDestAddr.sin_zero), '\0', 8);  // zero the rest of the struct
    
    // Send out the packet over the DOA OCD's socket.
    int numSent = sendto (mySockFD, packet.bytes(), packet.numBytes(), 0, (struct sockaddr *)& (tmpDestAddr), sizeof (struct sockaddr_in));
    return numSent;
}

/* Anytime a packet is received on DOA's socket, it is received in this method.*/
void DOAContext::receiveDOA(){
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"rcvPacket...\n");
	
	// Read the packet from the socket on which DOA OCD was listening.
    char recvBuf [ MAX_PACKET_BUF_LEN];
    struct sockaddr_in their_addr; 
    socklen_t addr_len = sizeof(struct sockaddr);
    int recvLen = recvfrom (    mySockFD, 
                            recvBuf, MAX_PACKET_BUF_LEN - 1 , 0,
			                (struct sockaddr *) &their_addr, &addr_len
                        );
    if (recvLen == -1) {
        PRINT_DEBUG ( DEBUG_LEVEL_WARNING,
                        "Error while receiving packet from %s: %s.\n",
                        inet_ntoa (their_addr.sin_addr),
                        strerror (errno)
                    );
        return;  
    
    } 
        
    PRINT_DEBUG ( DEBUG_LEVEL_VERBOSE,
                        "Received packet from %s:%d.\n",
                        inet_ntoa (their_addr.sin_addr), ntohs (their_addr.sin_port)
                    );
	//dns packet?
	if(memcmp(recvBuf,"dns",3)==0)
	{
		PRINT_DEBUG (DEBUG_LEVEL_SUPER,"its a dns packet...\n");
		eid reply;
		DOAStateInfo* state;
		memcpy(&state,recvBuf+3,4);
		if (memcmp(recvBuf+7, "DNSADDRESS_NOT_FOUND", 20)==0) {
		  PRINT_DEBUG (DEBUG_LEVEL_SUPER, "address not found\n");
		  state->receivedDNSReply(NULL);
		} 
		else {
		  PRINT_DEBUG (DEBUG_LEVEL_SUPER, "address found\n");		  
		  memcpy(&reply,recvBuf+7,20);
		  state->receivedDNSReply(&reply);
		}
		return;
	}	
	
	// dht packet?
	if (memcmp(recvBuf,"dht",3)==0)
	{
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "it's a dht packet. \n");
		DOAStateInfo* state;
		memcpy(&state,recvBuf+10,4);
		if (memcmp(recvBuf+3,"SUCCESS",7)==0) {
			PRINT_DEBUG (DEBUG_LEVEL_SUPER, "DHT resolution packet received.\n");		  
			state->receivedDHTReply(recvBuf+14);
		}
		else {
			PRINT_DEBUG (DEBUG_LEVEL_SUPER, "DHT lookup failed\n");		  
			state->receivedDHTReply(NULL); //signal failure
		}
		return;
	}			
	
	// if not dns or dht packet, make sure it's a DOA packet
	if(!isDOAPacket(recvBuf,recvLen)) return;
	
	// Must be a DOA data packet - process it
	DOAPacket packet = DOAPacket(recvBuf,recvLen);
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"its a data packet...\n");
	
	//is the packet destined to us?
	if(memcmp(packet.header->destination,(void*)&mEID,sizeof(eid))==0)
 	{
		PRINT_DEBUG (DEBUG_LEVEL_SUPER,"destination matches...\n");

		//if this packet is destined for someone else, forward it
		if (packet.routeLen > 1) { 
		
			PRINT_DEBUG(DEBUG_LEVEL_SUPER, "I am a delegate - final dest is not me... \n");
			struct sockaddr_in tmpDestAddr;
  			in_addr addr;
  			addr.s_addr = packet.header->destinationIPs[1];
  			tmpDestAddr.sin_addr = addr;
  			tmpDestAddr.sin_port = htons(DOA_PORT);
  
  			tmpDestAddr.sin_family = AF_INET;
  			memset(&(tmpDestAddr.sin_zero), '\0', 8);  // zero the rest of the struct
  
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"creating packet to forward...\n");
 			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"RecvLen = %d\n", recvLen);
 			
  			char fpacket[recvLen-sizeof(eid)-4];//* fpacket = new char[recvLen-sizeof(eid)-4];
  			int packet_offset = 0;
  			
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"about to copy doapacketheader\n");
  			memcpy(fpacket+packet_offset,recvBuf,sizeof(DOAPacketPreHeader));
  			packet_offset += sizeof(DOAPacketPreHeader);
  			
  			DOAPacketPreHeader* outheader = (DOAPacketPreHeader*) (&fpacket); 
  						
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"changing header length\n");
  			outheader->hdr_len = (packet.header->hdr_len*4-sizeof(eid)-4)/4;
  			
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"copying destination eid stack\n");
  			memcpy(fpacket+packet_offset,&(packet.header->destination[1]),(packet.routeLen-1)*sizeof(eid));
  			packet_offset += (packet.routeLen-1)*sizeof(eid);
  			
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"copying ip stack\n");
  			memcpy(fpacket+packet_offset, &(packet.header->destinationIPs[1]), (packet.routeLen-1)*4);
  			packet_offset += (packet.routeLen-1)*4;
  			
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"copying everything else\n");
			memcpy(fpacket+packet_offset, recvBuf+packet.header->hdr_len*4, recvLen-packet.header->hdr_len*4);
			packet_offset += recvLen-packet.header->hdr_len*4;
			
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Forwarding packet created.  Sending...\n");
  			int sendlen = recvLen-sizeof(eid)-4;
  			int fpacklen = sizeof(fpacket);
  			PRINT_DEBUG(DEBUG_LEVEL_SUPER, "lengths: %d, %d\n", sendlen, fpacklen);
  			int numSent = sendto (mySockFD,
					fpacket,  fpacklen, 
					0, 
					(struct sockaddr *)& (tmpDestAddr),
					sizeof (struct sockaddr_in));

			PRINT_DEBUG(DEBUG_LEVEL_SUPER,"forwarding packet sent.\n");
		}

       	// otherwise, pass the packet to the OCI since final destination is me
		else {
			DOAStateInfo* info = lookupRemoteHost(&(packet.header->source));
 		
 			// is the sender in our table of hosts?  If not...
 			if(info == NULL)
	 		{
		 		PRINT_DEBUG (DEBUG_LEVEL_SUPER,"new source! creating info...\n");
		        LayerHandle newOCDHandle = doaOCD->getNewOCDHandle();
		        PRINT_DEBUG (DEBUG_LEVEL_SUPER, "new handle for this ocdstateinfo is %d...\n", newOCDHandle);
		        
		        // create new DOAStateInfo for the connection to the incoming packet's source host
		        info = new DOAStateInfo (newOCDHandle, doaOCD->oci, doaOCD);			

		        // register this OCDState with the OCD.
		        doaOCD->putStateInfo (info->myHandle, info);
		       		
				// set the new host's eid to be the destination of the DOAStateInfo
				eid* addr = new eid;
				memcpy(addr,&(packet.header->source),sizeof(eid));
				info->route.push_back(addr); 
				// note: no DHT resolution done here, a call to send using this new state info
				// will set off the resolution cascade
				
				// add this new host to this context's [host -> DOAStateInfo] table		
		        addRemoteHost (info); 
	 		}
			PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Packet received on handle %d.\n", info->myHandle);
        	doaOCD->oci->receive (packet.payload(), doaOCD, info->myHandle);
        }		
 	}
 	else {
 		PRINT_DEBUG (DEBUG_LEVEL_SUPER,"destination does NOT match - throwing away...\n");
 	}
}

void DOAContext::addRemoteHost (DOAStateInfo* currOCDState) {
    eid* addr = new eid;
    memcpy(addr,currOCDState->route[currOCDState->route.size()-1],sizeof(eid));
    addrToState.insert ( pair<eid*, DOAStateInfo*>(addr, currOCDState));
}

DOAStateInfo* DOAContext::lookupRemoteHost (eid* addr) {
    map<eid*, DOAStateInfo*>::iterator itr = addrToState.find (addr);
    
    while (itr != addrToState.end()) {
 	   return itr->second;
    }
    return NULL;
}

eid* DOAContext::myEID(){
	//TODO fix dummy
	return &mEID;
}
