#ifndef DOAPACKET_H_
#define DOAPACKET_H_
/* 
 * DOAPacket.h
 * Defines the structure of DOA overlay packet and provides utility functions for 
 * construction and manipulation of them.
 * 
 * @author levpopov, djwendel, evhan55
 * Code adapted from doa.hh and SampleOCDPacket.h
 */


#include <inttypes.h>
#include <vector>
#include <assert.h>

using namespace std;

#include "DOAContext.h"
#include "DOAStateInfo.h"
#include "DOAGlobals.h"

/* The first portion of the actual DOA header.
 * The destination EID/IP stacks are kept in separate structures
 * because they have unknown sizes.
 */
typedef struct {
	unsigned version: 3; 
	unsigned hdr_len: 5; 
	uint8_t protocol;
	uint16_t length;
	eid source;
	eid* destination;
	uint32_t* destinationIPs;
} DOAPacketPreHeader;

/* TODO Returns non-0 iff data is a valid DOA packet */
bool isDOAPacket(char* data,int len){
	return (len >= sizeof(DOAPacketPreHeader)+24);
}

/* DOA Packet encapsulation with utility methods */
class DOAPacket{
	DOAPacketPreHeader* header;

	//number of EIDs in destination stack
	int routeLen;

	//number of IPs in destinationIPs stack
	int IPRouteLen;

	char* pBytes;
	int numPBytes;

public:	
	/* Construct DOAPacket from outgoing IP packet */
	DOAPacket(DOAStateInfo* info, DOAContext* context, char* data, int len);
	/* Construct DOAPAcket from incoming DOA packet */
	DOAPacket(char* data,int len);

	virtual ~DOAPacket(); 

	/* returns the actual packet buffer, ready to be sent */
	char* bytes();	
	
	/* returns the size of the actual packet buffer, corresponds to bytes() */
	int numBytes();
	
	char* payload();
	int payloadLen();
	
	friend class DOAContext;
};

DOAPacket::DOAPacket(DOAStateInfo* info, DOAContext* context, char* data, int len){
	
	int size = sizeof(DOAPacketPreHeader)+sizeof(eid)*(info->route.size())+4*(info->ip_hints.size())+len;
	assert((pBytes = (char*)malloc(size)) != NULL);
	numPBytes = size;
	
	header = (DOAPacketPreHeader*)pBytes;
	header->version = DOA_VERSION;
	/*TODO check that we can actually fit that much into header */
	header->hdr_len = (size-len)/4; //in words so we divide by 4
	header->protocol = DOA_PROTOCOL;
	header->length = size;
	memcpy((void*)&(header->source),(void*)context->myEID(),sizeof(eid));
	header->destination = (eid*)(header+1);	

	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Beginning to copy EID stack...\n");

	//copy over destination route
	routeLen = info->route.size();
	PRINT_DEBUG(DEBUG_LEVEL_SUPER,"routeLen: %d, size: %d...\n", routeLen,size);
	for(int i=0;i<info->route.size();i++)
		memcpy((char*)pBytes+sizeof(DOAPacketPreHeader)+sizeof(eid)*i,(info->route[i]),sizeof(eid));

	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Beginning to copy IP stack...\n");	

	//copy over IPdestination route
	if(&(info->ip_hints[0]) == NULL)
		PRINT_DEBUG(DEBUG_LEVEL_SUPER,"iphints is NULL!!\n");
		
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "ip_hints.size() = %d...\n", info->ip_hints.size());

	IPRouteLen = info->ip_hints.size();
	for(int i=0;i<info->ip_hints.size();i++)
		memcpy((char*)pBytes+sizeof(DOAPacketPreHeader)+sizeof(eid)*routeLen+4*i,&(info->ip_hints[i]),4);

	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Beginning to copy the rest...\n");

	//copy the rest
	memcpy((char*)pBytes+header->hdr_len*4,data,len);
	
	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Preparing to print final message...\n");
	
	char src[21];
	char dst[21];
	memcpy(src,(void*)&(header->source),20);
	memcpy(dst,(void*)(header->destination),20);
	src[20]=0;
	dst[20]=0;
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"outgoing DOAPACKET from %s to %s...\n",src,dst);	
}

DOAPacket::DOAPacket(char* data, int len){
	//TODO add packet integrity verification
	assert((pBytes = (char*)malloc(len)) != NULL);
	memcpy(pBytes,data,len);
	header = (DOAPacketPreHeader*)pBytes;
	int restLength = (header->hdr_len*4-sizeof(DOAPacketPreHeader));
	
	// read out destination EID stack
	routeLen = restLength/(sizeof(eid)+4);
	header->destination = new eid[routeLen];
	for(int i=0;i<routeLen;i++)
		memcpy(&(header->destination[i]),data+sizeof(DOAPacketPreHeader)+sizeof(eid)*i,sizeof(eid));
		
	// read out destination IP stack
	IPRouteLen = routeLen;
	header->destinationIPs = new uint32_t[IPRouteLen];
	for(int i=0;i<IPRouteLen;i++)
		memcpy(&(header->destinationIPs[i]),data+sizeof(DOAPacketPreHeader)+sizeof(eid)*routeLen+4*i,4);
	
	numPBytes = len;
	
	char src[21];
	char dst[21];
	memcpy(src,(void*)&(header->source),20);
	memcpy(dst,(void*)(header->destination),20);
	src[20]=0;
	dst[20]=0;
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"incoming DOAPACKET from %s to %s...\n",src,dst);	
}

char* DOAPacket::bytes(){
	return pBytes;
}

int DOAPacket::numBytes(){
	return numPBytes;
}

char* DOAPacket::payload(){
	return pBytes+header->hdr_len*4;
}

int DOAPacket::payloadLen(){
	return numPBytes-header->hdr_len*4;
}

DOAPacket::~DOAPacket(){
	if(pBytes) free(pBytes);
}
#endif /*DOAPACKET_H_*/
