
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "OCD_include.h"

#include "DOAStateInfo.h"
#include "DOAOCD.h"
#include "DOAContext.h"


DOAStateInfo::DOAStateInfo(
			LayerHandle myHandle, OCI* oci, DOAOCD *overlay
			/* OCD specific parameters */
		)
            : OCDStateInfo (myHandle, oci, overlay) {
	PRINT_DEBUG (DEBUG_LEVEL_SUPER,"creating doaStateInfo...\n");
	
    // OCD specific initialization
	// zero out the destination fields
	memset(&dest_eid,'\0',EID_SIZE);
	memset(&(dest_ip.s_addr),'\0',4);

	// set this StateInfo to be "not ready" until after DNS/DHT resolution
	readyFlag = false;
	resolvingFlag = false;

	// make room for packets that we have to queue until the StateInfo IS ready
	packets = new QueuedPacket*[100];
	numQPackets = 0;
} 

/**
  * This function is called when the DNS reply for the end point name of this OCD state
  * has been received.
  *
  * NOTE - in our current implementation we don't design for DNS security since DNS 
  * resolution is not the focus of DOA
  */
void DOAStateInfo::receivedDNSReply (eid* address) {
	PRINT_DEBUG (DEBUG_LEVEL_SUPER, "rcvDNSReply...\n");
    if (address == 0) {
        // DNS resolution failed.
        // Tell OCI that setup failed.
        notifySetupFailure();
    } else {
        PRINT_DEBUG (DEBUG_LEVEL_SUPER,
                    "DNS resolved %s. Continuing tunnel setup...\n",
                    this->endPointName
                );
                
        // add the resolved eid to the end of the eid stack.  It's our final destination.
        eid* dest = new eid;
        memcpy(dest,address,sizeof(eid));
        vector<eid*>::iterator routefront = route.begin();
        this->route.insert(routefront, dest);
        dest_eid = *dest;
        
        // resolve this route to get the first delegate's IP address
        checkAndResolve();
        
        // register in eid->stateinfo hashtable
        ((DOAOCD*)this->ocd)->context->addRemoteHost (this);
    }
}


/**
 * This method checks to see if the next-hop IP address has been discovered yet.
 * If it hasn't, it sends a DHT request to resolve the first EID on the route.
 */
void DOAStateInfo::checkAndResolve() {

	// if the IP destination as still empty, we have to keep resolving
	if (memcmp(&(dest_ip.s_addr),"\0\0\0\0",4)==0) {
		resolveEID(route[0], this); //resolve the first eid in the route
		readyFlag = false;
		resolvingFlag = true;
	}
	
	//otherwise, we've completed the address resolution and tunnel setup, so notify,
	// set readyFlag to true, and send all queued up messages waiting to be sent
	else {
		readyFlag = true;
		resolvingFlag = false;
		notifySetupSuccess();
		sendQueued();
	}
}


/**
 * This method is called whenever a DHT lookup reply is received.  The method parses
 * the reply to get out the erecord data.  It then calls checkAndResolve which will
 * either notify the OCI of success or continue resolving the EID stack.
 *
 * TODO: use an erecord data type to make things cleaner
 * TODO: add verification (via self-certs) for all incoming DHT resolve packets.
 */
void DOAStateInfo::receivedDHTReply(char* erec_string) {

	/* First thing to check is that the erecord is signed by the public
	 * key, and that the public key hashes to EID we were looking for.
	 * If not, discard.
	 *
	 * (These security features currently not implemented)
	 */

	// first, only run this method if we're waiting for a DHT reply
	if (memcmp(&(dest_ip.s_addr),"\0\0\0\0",4)!=0) {
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "DHT reply received, but setup already done.  Discarding.\n");
		return;
	}
	
	// if no entry found
	if (erec_string == NULL) {
		notifySetupFailure(); // can't set up tunnel.
		return;
	}

	// now start walking down the packet, reading useful information along the way
	int packet_offset = 0;
	eid erec_eid;
	memcpy((char*)(&erec_eid), erec_string+packet_offset, 20);
	packet_offset += 20;

	// we're only interested in the erecord we requested, so check to make sure we got that.
	if (route[0] != NULL && memcmp(&erec_eid,route[0],20)!=0) {
		char exp[21];
		char rec[21];
		memcpy(exp, route[0], 20);
		exp[20] = 0;
		memcpy(rec, &erec_eid, 20);
		rec[20] = 0;		
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "DHT reply received for wrong eid.\nExpected: %s\nReceived: %s. Discarding. \n", exp, rec);
		checkAndResolve();
		return;
	}

	// get the number of delegate EIDs (currently only 0 or 1 are supported)
	uint16_t numEIDs = 0;
	memcpy((char*)(&numEIDs), erec_string+packet_offset, 2);
	packet_offset += 2;

	PRINT_DEBUG(DEBUG_LEVEL_SUPER, "numEIDs = %d\n", numEIDs);
	
	// if numEIDs == 0, erecord contains one IP address (our next hop) and no EIDs
	if (numEIDs == 0) {
		// grab IP address from erecord
		memcpy((char*)(&(dest_ip.s_addr)), erec_string+packet_offset, 4);
		packet_offset += 4;		
	}
	// otherwise, the erecord contains an EID stack and no IP delegates
	else {
		// grab EID stack from erecord
		eid* tempEID;
		for (int i=0; i < numEIDs; i++) {
			tempEID = new eid;
			memcpy((char*)tempEID, erec_string+packet_offset, 20);
			packet_offset += 20;
	        vector<eid*>::iterator routefront = route.begin();
    	    route.insert(routefront, tempEID);
		}
		char eidprint[21];
		memcpy(eidprint, tempEID, 20);
		eidprint[21] = '\0';
		PRINT_DEBUG(DEBUG_LEVEL_SUPER,"Grabbed EID %s from stack.\n", eidprint);
	}
	
	// grab hint IP address and put it in the front of the IP stack
	uint32_t hintIP;
	memcpy((char*)(&hintIP), erec_string+packet_offset, 4);
	packet_offset += 4;		
	vector<uint32_t>::iterator hintfront = ip_hints.begin();
	ip_hints.insert(hintfront,hintIP);
	
	// we've resolved one level, so check if we can send now.
	checkAndResolve();	
}

/**
 * This method returns true only if the next hop IP address has been found
 * (e.g. after all of the DHT lookups have returned)
 */
bool DOAStateInfo::isReady() {
	return readyFlag;
}

/**
 * This method returns true if this StateInfo is in the process of 
 * doing a DHT resolution lookup.  Otherwise returns false
 */
bool DOAStateInfo::isResolving() {
	return resolvingFlag;
}

/**
 * If this state isn't ready to send packets yet (DHT replies haven't returned),
 * the messages will be queued up instead, waiting for the state to be ready.
 */
void DOAStateInfo::enqueuePacket(char* data, int len) {
	QueuedPacket* q = new QueuedPacket;
	
	// copy everything because the higher layers think this packet is
	// sent, so they could clobber the buffer.
	char* data_copy = new char[len];
	memcpy(data_copy, data, len);
	q->len = len;
	q->data = data_copy;
	
	if (numQPackets < 100) {
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Enqueing packet #%d...\n", numQPackets);
		packets[numQPackets] = q;
		numQPackets++;
	}
	else {
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "No room to enqueue next packet.  Dropping it.\n");
	}
}


/**
 * This method sends all of the packets that have been waiting to be sent while
 * the DHT lookups were happening.
 */
void DOAStateInfo::sendQueued() {
	// send all pending packets, and delete them from the queue
	for (int i = 0; i<numQPackets; i++) {
		PRINT_DEBUG(DEBUG_LEVEL_SUPER, "Sending queued packet #%d.\n", i);
		
		// send the packet
		((DOAOCD*)this->ocd)->context->sendDOA(this, packets[i]->data, packets[i]->len);
		
		// and delete it from the queue
		delete(packets[i]->data);
		delete(packets[i]);		
	}
}


DOAStateInfo::~DOAStateInfo(){
	for(int i=0;i<route.size();i++)
	{
		delete(route[i]);
	}
	route.clear();
	ip_hints.clear();
	for (int j = 0; j<numQPackets; j++) {
		delete(packets[j]->data);
		delete(packets[j]);
	}
	delete(packets);
}
